from __future__ import annotations
from typing import TYPE_CHECKING, Dict, Tuple, List, Sequence


if TYPE_CHECKING:
    from pybi.core.components.reactiveComponent import EChartDatasetInfo


m_merge_keys = set(["tooltip", "legend", "title", "grid"])


class BaseChart:
    def __init__(self) -> None:
        self.__base_opt = {
            "tooltip": {},
            "legend": {},
            "series": [],
            "title": {},
            "grid": {"containLabel": True},
        }
        self.__merge_opt = {}

    def merge(self, options: Dict):
        self.__merge_opt = options
        return self

    def get_options(self):
        return {
            **self.__base_opt,
            **{k: v for k, v in self.__merge_opt.items() if k in m_merge_keys},
        }

    def set_title(self, text: str):
        opt_title: Dict = self.__base_opt["title"]
        opt_title["text"] = text

        return self

    def get_options_infos(self) -> Tuple[Dict, Sequence[EChartDatasetInfo]]:
        raise NotImplementedError



from pybi.core.components.component import Component
from .componentTag import ComponentTag


class TextComponent(Component):
    def __init__(self, content: str) -> None:
        super().__init__(ComponentTag.Text)
        self.content = content


class UploadComponent(Component):
    def __init__(self) -> None:
        super().__init__(ComponentTag.Upload)



from __future__ import annotations
from .line import LineChart
from .bar import BarChart
from .pie import PieChart
from .scatter import ScatterChart
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


__all__ = ["easy_echarts"]


class EasyEChartsMeta:
    def make_line(
        self,
        data: DataSourceTable,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="avg",
    ):
        return LineChart(data, x, y, color, agg)

    def make_bar(
        self,
        data: DataSourceTable,
        *,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="avg",
    ):
        return BarChart(data, x, y, color, agg)

    def make_pie(
        self,
        data: DataSourceTable,
        *,
        name: str,
        value: str,
        agg="avg",
    ):
        return PieChart(data, name, value, agg)

    def make_scatter(
        self,
        data: DataSourceTable,
        *,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="avg",
    ):
        return ScatterChart(data, x, y, color, agg)


easy_echarts = EasyEChartsMeta()



from __future__ import annotations
from typing import TYPE_CHECKING, List, Union


from pybi.core.components import ComponentTag
from pybi.core.sql import Sql
from .base import SingleReactiveComponent


if TYPE_CHECKING:
    pass


class Table(SingleReactiveComponent):
    def __init__(
        self,
        sql: Sql,
    ) -> None:
        super().__init__(ComponentTag.Table, sql)



from __future__ import annotations
from typing import TYPE_CHECKING, Union, List, Optional

from .componentTag import ComponentTag
from .component import Component

if TYPE_CHECKING:
    from pybi.app import App


class ContainerComponent(Component):
    def __init__(self, tag: ComponentTag, appHost: Optional[App] = None) -> None:
        super().__init__(tag)

        # TODO: maybe use weak ref?
        self._appHost = appHost

        self.children: List[Component] = []

    def _add_children(self, stat: Component):
        self.children.append(stat)
        return self

    def __enter__(self):
        if self._appHost:
            self._appHost._with_temp_host_stack.append(self)
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        if self._appHost:
            self._appHost._with_temp_host_stack.pop()


class BoxComponent(ContainerComponent):
    def __init__(self, appHost: Optional[App] = None) -> None:
        super().__init__(ComponentTag.Box, appHost)


class FlowBoxComponent(ContainerComponent):
    def __init__(self, appHost: Optional[App] = None) -> None:
        super().__init__(ComponentTag.FlowBox, appHost)

    def __get_item__(self, idx: int):
        return self.children[idx]


class GridBoxComponent(ContainerComponent):
    def __init__(self, areas: List[List[str]], appHost: Optional[App] = None) -> None:
        super().__init__(ComponentTag.GridBox, appHost)
        self.__areas = areas
        self.__columns_sizes: List[str] = []
        self.__rows_sizes: List[str] = []

    def set_columns_sizes(self, sizes: List[Union[str, int]]):
        """
        >>> set_columns_sizes([1,2,2])
        >>> set_columns_sizes(['150px','1fr','2fr'])
        >>> set_columns_sizes(['150px','1fr','minmax(100px, 1fr)'])
        """
        self.__columns_sizes = [f"{s}fr" if isinstance(s, int) else s for s in sizes]
        return self

    def set_rows_sizes(self, sizes: List[Union[str, int]]):
        """
        >>> set_rows_sizes([1,2,2])
        >>> set_rows_sizes(['150px','1fr','2fr'])
        >>> set_rows_sizes(['150px','1fr','minmax(100px, 1fr)'])
        """
        self.__rows_sizes = [f"{s}fr" if isinstance(s, int) else s for s in sizes]
        return self

    def __get_item__(self, idx: int):
        return self.children[idx]

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["areas"] = GridBoxComponent.areas_array2str(self.__areas)

        areas_cols_len = max(map(len, self.__areas))
        cols_size = GridBoxComponent.padded_grid_template(
            self.__columns_sizes, areas_cols_len
        )

        areas_rows_len = len(self.__areas)
        rows_size = GridBoxComponent.padded_grid_template(
            self.__rows_sizes, areas_rows_len
        )

        data["gridTemplateColumns"] = " ".join(cols_size)
        data["gridTemplateRows"] = " ".join(rows_size)

        return data

    @staticmethod
    def padded_grid_template(sizes: List[str], real_size: int):
        """
        >>> sizes = ['1rf']
        >>> real_size = 3
        >>> padded_grid_template(sizes,real_size)
        >>> ['1rf','auto','auto']
        """
        diff_len = real_size - len(sizes)
        sizes = sizes.copy()

        if diff_len > 0:
            sizes.extend(["auto"] * diff_len)
        return sizes

    @staticmethod
    def areas_array2str(areas_array: List[List[str]]):
        """
        >>> input = [
            ["sc1", "sc2"],
            ["sc3"],
            ["table"] * 4
        ]
        >>> areas_array2str(input)
        >>> '"sc1 sc2 . ." "sc3 . . ." "table table table table"'
        """
        max_len = max(map(len, areas_array))

        fix_empty = (
            [*line, *(["."] * (max_len - len(line)))] if len(line) < max_len else line
            for line in areas_array
        )

        line2str = (f'"{" ".join(line)}"' for line in fix_empty)
        return " ".join(line2str)

    @staticmethod
    def areas_str2array(areas: str) -> List[List[str]]:
        """
        >>> input='''
            sc1 sc2
            sc3
            table table table table
        '''
        >>> areas_str2array(input)
        >>> [
            ["sc1", "sc2"],
            ["sc3"],
            ["table", "table", "table", "table"]
        ]
        """
        pass

        lines = (line.strip() for line in areas.splitlines())
        remove_empty_rows = (line for line in lines if len(line) > 0)
        splie_space = (line.split() for line in remove_empty_rows)
        return list(splie_space)


class ColBoxComponent(ContainerComponent):
    def __init__(
        self, spec: Optional[List[int]] = None, appHost: Optional[App] = None
    ) -> None:
        super().__init__(ComponentTag.ColBox, appHost)

        if spec is None:
            spec = [1, 1]

        self.spec = spec

        assert self._appHost is not None, "self._appHost must be app instance"

    def __getitem__(self, idx: int):
        return self.children[idx]


class TabsComponent(ContainerComponent):
    def __init__(
        self, names: List[str], mode="narrowing", appHost: Optional[App] = None
    ) -> None:
        """
        mode: 'fullWidth' | 'narrowing'
        """
        super().__init__(ComponentTag.Tabs, appHost)
        self.names = names
        self.mode = mode

        for _ in range(len(self.names)):
            self._add_children(BoxComponent(appHost))

    def __getitem__(self, idx: int):
        res = self.children[idx]
        assert isinstance(res, ContainerComponent)
        return res



from __future__ import annotations
from .base import BaseChart
import pybi as pbi

from typing import TYPE_CHECKING, Optional
from pybi.core.components.reactiveComponent import EChartDatasetInfo

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


class BarChart(BaseChart):
    def __init__(
        self,
        data: DataSourceTable,
        x: str,
        y: str,
        color: Optional[str] = None,
        agg="avg",
    ):
        super().__init__()
        self.data = data
        self.x = x
        self.y = y
        self.color = color
        self.agg = agg
        self._reverse_axis = False
        self._series_configs = {}

    def reverse_axis(self):
        self._reverse_axis = True
        return self

    def get_options_infos(self):
        opts = super().get_options()

        opt_data = None
        if self.color:
            opt_data = self.data.to_pivot(
                row=self.x, column=self.color, cell=self.y, agg=self.agg
            )
        else:
            opt_data = pbi.set_dataView(
                f'select `{self.x}`,{self.agg}(`{self.y}`) as "{self.y}" from {self.data} group by `{self.x}`'
            )

        series_config = {"type": "bar"}
        series_config.update(self._series_configs)

        ds_info = EChartDatasetInfo(series_config, "dataset[0]", opt_data.source_name)

        valueAxisConfig = {"type": "value"}
        catAxisConfig = {"type": "category"}

        if self._reverse_axis:
            opts["xAxis"] = [valueAxisConfig]
            opts["yAxis"] = [catAxisConfig]
        else:
            opts["xAxis"] = [catAxisConfig]
            opts["yAxis"] = [valueAxisConfig]

        return opts, [ds_info]



from __future__ import annotations
from .base import BaseChart
import pybi as pbi

from typing import TYPE_CHECKING, Optional
from pybi.core.components.reactiveComponent import EChartDatasetInfo

if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


class PieChart(BaseChart):
    def __init__(
        self,
        data: DataSourceTable,
        name: str,
        value: str,
        agg="avg",
    ):
        super().__init__()
        self.data = data
        self.name = name
        self.value = value
        self.agg = agg
        self._series_configs = {}

    def radius(self, value=["40%", "70%"]):
        """
        设置圆环图颜色。
        参考资料:https://echarts.apache.org/zh/option.html#series-pie.radius
        """
        self._series_configs["radius"] = value
        self._series_configs["avoidLabelOverlap"] = False
        self._series_configs["label"] = {"show": False, "position": "center"}
        self._series_configs["labelLine"] = {"show": False}
        self._series_configs["emphasis"] = {
            "label": {"show": True, "fontSize": "2em", "fontWeight": "bold"}
        }

        return self

    def get_options_infos(self):
        opts = super().get_options()

        opt_data = pbi.set_dataView(
            f"select `{self.name}`,{self.agg}(`{self.value}`) from {self.data} group by `{self.name}`"
        )

        series_config = {
            "type": "pie",
            "radius": "50%",
            "emphasis": {
                "itemStyle": {
                    "shadowBlur": 10,
                    "shadowOffsetX": 0,
                    "shadowColor": "rgba(0, 0, 0, 0.5)",
                }
            },
        }

        series_config.update(self._series_configs)

        ds_info = EChartDatasetInfo(series_config, "dataset[0]", opt_data.source_name)

        opts["xAxis"] = None

        opts["yAxis"] = None

        opts["tooltip"] = {"trigger": "item"}

        return opts, [ds_info]



from __future__ import annotations
from typing import TYPE_CHECKING, List, Optional, Dict, Set, Any, Union

from pybi.utils.data_gen import Jsonable

from pybi.core.components import ComponentTag
from pybi.core.sql import Sql
from .base import ReactiveComponent
from pybi.core.dataSource import DataSourceView, DataSourceTable


if TYPE_CHECKING:
    from pybi.core.dataSource import DataSourceTable


class EChartSqlInfo(Jsonable):
    def __init__(self, seriesConfig: Dict, path: str, sql: Sql, chartType: str) -> None:
        self.path = path
        self._sql = sql
        self.chartType = chartType
        self.seriesConfig = seriesConfig

    def _to_json_dict(self):
        data = super()._to_json_dict()
        sql_data = self._sql._to_json_dict()

        data = {**data, **sql_data}
        return data


class EChartDatasetInfo(Jsonable):
    def __init__(self, seriesConfig: Dict, path: str, dataset: str) -> None:
        self.path = path
        self.dataset = dataset
        self.seriesConfig = seriesConfig


class EChartUpdateInfo(Jsonable):
    def __init__(
        self,
        action_type: str,
        value_type: str,
        table: str,
        field: str,
    ) -> None:
        """
        action_type : Literal["hover", "click"]
        value_type: Literal["x", "y", "value"]
        """
        super().__init__()
        self.actionType = action_type
        self.valueType = value_type
        self.table = table
        self.field = field


class EChart(ReactiveComponent):
    def __init__(self, options: Dict) -> None:

        super().__init__(ComponentTag.EChart)

        self._chart_mappings = {}
        self.options = options
        self._sqlInfos: List[EChartDatasetInfo] = []
        self._chartUpdateInfos: List[EChartUpdateInfo] = []
        self.height = "100%"

    def set_height(self, value: str):
        """
        15em:15字体大小
        300px:300像素
        """
        self.height = value
        return self

    def add_sql_path(self, info: EChartDatasetInfo):
        """
        add_sql_path('xAxis.data',"select distinct name from data")
        """
        self._sqlInfos.append(info)
        return self

    def hover_filter(
        self, value_type: str, table: Union[str, DataSourceTable], field: str
    ):
        """
        value_type: Literal["x", "y", "value"]
        """
        if isinstance(table, DataSourceTable):
            table = table.source_name

        self._chartUpdateInfos.append(
            EChartUpdateInfo("hover", value_type, table, field)
        )
        return self

    def _to_json_dict(self):
        self.options["series"] = []
        # EChart.remove_sql_from_option(self.options, self._sqlInfos)

        EChart.remove_legend_empty_data(self.options)
        EChart.remove_axis_empty_data(self.options)

        data = super()._to_json_dict()
        data["mapping"] = self._chart_mappings

        data["sqlInfos"] = self._sqlInfos
        data["updateInfos"] = self._chartUpdateInfos

        return data

    @staticmethod
    def remove_legend_empty_data(options: Dict):
        target = options
        if "legend" in target:
            target = options["legend"]
            if isinstance(target, list):
                target = target[0]
                if (
                    "data" in target
                    and isinstance(target["data"], list)
                    and len(target["data"]) == 0
                    or (not target["data"][0])
                ):
                    target["data"] = None

    @staticmethod
    def remove_axis_empty_data(options: Dict):
        def remove_fn(name: str):
            target = options
            if name not in target or target[name] is None:
                return

            if "data" in target[name][0]:
                axis = target[name][0]
                target = axis["data"]
                if (
                    isinstance(target, list)
                    and len(target) == 0
                    or (not target[0] == "")
                ):
                    axis["data"] = None
                    axis["type"] = "category"

        remove_fn("xAxis")
        remove_fn("yAxis")

    @staticmethod
    def extract_infos_from_option(opt: Dict):

        yield from EChart._extract_series_infos_from_option_(opt)
        return
        stack = [("", opt)]

        while len(stack) > 0:

            path, target = stack.pop()

            for key, value in target.items():
                if key != "series":
                    cur_path = f"{path}.{key}"

                    if isinstance(value, dict):
                        stack.append((cur_path, value))
                    elif isinstance(value, list):
                        for idx, v in enumerate(value):
                            if isinstance(v, dict):
                                stack.append((f"{cur_path}[{idx}]", v))
                    elif isinstance(value, (Sql, DataSourceView)):

                        if isinstance(value, DataSourceView):
                            value = Sql(value._to_sql())

                        yield EChartSqlInfo(cur_path[1:], value, "other")

    @staticmethod
    def _extract_series_infos_from_option_(opt: Dict):
        if "series" not in opt:
            return

        def extract_series(series: Dict, idx=-1):
            series_data = series["data"]
            if series["type"] in ["line", "pie", "bar"]:
                if isinstance(series_data, (Sql, DataSourceView)):
                    if isinstance(series_data, DataSourceView):
                        series_data = Sql(series_data._to_sql())

                    path = "series"
                    if idx > -1:
                        path = f"{path}[{idx}].data"

                    series["pybiFlag"] = None

                    yield EChartSqlInfo(series, path, series_data, series["type"])

        series = opt["series"]
        if isinstance(series, list):
            for idx, s in enumerate(series):
                yield from extract_series(s, idx)
        else:
            yield from extract_series(series)

    @staticmethod
    def _extract_dataset_infos_from_option_(opt: Dict):
        if "dataset" not in opt:
            return

        series = []
        if "series" in opt:
            series = opt["series"]

        def extract_ds(ds: Dict, series: List[Dict], idx=-1):
            if "source" not in ds:
                return

            source = ds["source"]

            if isinstance(source, (DataSourceTable)):

                path = "dataset"
                if idx > -1:
                    path = f"{path}[{idx}]"

                # series["pybiFlag"] = None

                if "data" in series[0]:
                    series[0]["data"] = None

                yield EChartDatasetInfo(series[0], path, source.source_name)

        dataset = opt["dataset"]
        assert isinstance(dataset, list)

        for idx, ds in enumerate(dataset):
            yield from extract_ds(ds, series, idx)

    @staticmethod
    def remove_sql_from_option(opt: Dict, infos: List[EChartDatasetInfo]):

        filterSeries = [series for series in opt["series"] if not "pybiFlag" in series]

        opt["series"] = filterSeries

        # for info in infos:
        #     paths = info.path.split(".")
        #     dictUtils.set_by_paths(paths, opt, None)



from __future__ import annotations
from pybi.utils.data_gen import get_global_id, Jsonable, random_dv_name
import pandas as pd
from enum import Enum
from typing import Dict, List, Any, Callable, Optional, TYPE_CHECKING, Union

from pybi.core.sql import Sql

if TYPE_CHECKING:
    from pybi.app import App


class DataViewType(Enum):
    Sql = "sql"
    Pivot = "pivot"


class DataViewBase(Jsonable):
    def __init__(self, name: str, type: DataViewType) -> None:
        super().__init__()
        self.name = name
        self.type = type
        self.__excludeLinkages: List[str] = []

    def exclude_source(self, name: str):
        self.__excludeLinkages.append(name)
        return self

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["excludeLinkages"] = self.__excludeLinkages
        return data


class DataView(DataViewBase):
    def __init__(self, name: str, sql: Sql) -> None:
        super().__init__(name, DataViewType.Sql)
        self._sql = sql

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["sql"] = self._sql.sql
        return data


class PivotDataView(DataViewBase):
    def __init__(
        self,
        name: str,
        source: str,
        row: str,
        column: str,
        cell: str,
        agg="min",
        excludeRowFields=False,
    ) -> None:
        super().__init__(name, DataViewType.Pivot)
        self.source = source
        self._row = row
        self._column = column
        self._cell = cell
        self._agg = agg
        self._excludeRowFields = excludeRowFields

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["pivotOptions"] = {
            "row": self._row,
            "column": self._column,
            "cell": self._cell,
            "agg": self._agg,
            "excludeRowFields": self._excludeRowFields,
        }
        return data


class DataSource(Jsonable):
    def __init__(self, name: str, data: pd.DataFrame) -> None:
        super().__init__()
        self._data = data
        self.name = name

    def __str__(self) -> str:
        return self.name


class DataSourceView:
    def __init__(self) -> None:
        pass

    def _to_sql(self):
        raise NotImplementedError()


class DataSourceField(DataSourceView):
    def __init__(self, source_name: str, name: str) -> None:
        super().__init__()
        self.source_name = source_name
        self.name = name

    def _get_sql_field_name(self):
        return f"`{self.name}`"

    def __str__(self) -> str:
        return str(Sql(self._to_sql()))

    def _to_sql(self):
        return f"select {self._get_sql_field_name()} from {self.source_name}"


class PyechartsPieItem(DataSourceView):
    def __init__(self, ds_table: DataSourceTable) -> None:
        super().__init__()
        self.ds_table = ds_table

    def __getitem__(self, key):
        from pyecharts import options as opts

        if key > 0:
            raise StopIteration()

        return opts.PieItem("", 0)

    def _to_sql(self):

        if len(self.ds_table.columns) < 2:
            raise Exception("pie data must have 2 fields")

        name = f"{self.ds_table.columns[0]} as name"
        value = f"{self.ds_table.columns[1]} as value"

        return f"select {name} , {value} from ({self.ds_table._to_sql()})"


class PyechartsLineItem(DataSourceView):
    def __init__(self, ds_table: DataSourceTable) -> None:
        super().__init__()
        self.ds_table = ds_table

    def __getitem__(self, key):
        from pyecharts import options as opts

        if key > 0:
            raise StopIteration()

        return opts.LineItem("", 0)

    def _to_sql(self):

        if len(self.ds_table.columns) < 3:
            raise Exception("pie data must have 3 fields")

        series_name = f"{self.ds_table.columns[0]} as sreies"
        name = f"{self.ds_table.columns[1]} as name"
        value = f"{self.ds_table.columns[2]} as value"

        return f"select {series_name},{name} , {value} from ({self.ds_table._to_sql()})"


class DataSourceTable(DataSourceView):
    def __init__(
        self,
        source_name: str,
        columns: List[str],
        *,
        host: App,
    ) -> None:
        super().__init__()
        self.source_name = source_name
        self.columns = columns
        self.__host = host

    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration

    def __getitem__(self, field: Union[str, List[str]]):
        if isinstance(field, str):
            return DataSourceField(self.source_name, field)
        return DataSourceTable(self.source_name, field, host=self.__host)

    def to_pivot(
        self,
        row: str,
        column: str,
        cell: str,
        agg="min",
        exclude_source: Optional[List[DataSourceTable]] = None,
        excludeRowFields: bool = False,
    ):
        """
        agg : sqlite aggregate function,like 'min','max','sum','round(avg(${}),2)'
            default:'min'
        """
        return self.__host.set_pivot_dataView(
            self.source_name, row, column, cell, agg, exclude_source, excludeRowFields
        )

    def to_pyecharts_pie_items(self) -> Any:
        return PyechartsPieItem(self)

    def to_pyecharts_line_items(self) -> Any:
        return PyechartsLineItem(self)

    def get_sql_fields(self):
        """
        >>> *
        >>> a,b,c
        """
        cols = (f"`{c}`" for c in self.columns)
        return "*" if len(self.columns) == 0 else ",".join(cols)

    def _to_sql(self):
        fields = self.get_sql_fields()

        return f"select {fields} from {self.source_name}"

    def __str__(self) -> str:
        return self.source_name
